;; Generated package description from jade-mode.el  -*- no-byte-compile: t -*-
(define-package "jade-mode" "1.0.1" "Major mode for editing .jade files" 'nil :commit "dad17dc86c93401646802a639a98dd2ec875db6f" :keywords '("languages") :url "https://github.com/brianc/jade-mode")
